from gaphor import UML
from gaphor.tests import TestCase
from gaphor.diagram.components import NodeItem, ComponentItem, ArtifactItem
from gaphor.diagram.usecases import UseCaseItem


class NodesGroupTestCase(TestCase):
    """
    Nodes grouping tests.
    """

    def test_grouping(self):
        """Test node within another node composition
        """
        n1 = self.create(NodeItem, UML.Node)
        n2 = self.create(NodeItem, UML.Node)

        self.group(n1, n2)

        assert n2.subject in n1.subject.nestedNode
        assert n1.subject not in n2.subject.nestedNode

    def test_ungrouping(self):
        """Test decomposition of component from node
        """
        n1 = self.create(NodeItem, UML.Node)
        n2 = self.create(NodeItem, UML.Node)

        self.group(n1, n2)
        self.ungroup(n1, n2)

        assert n2.subject not in n1.subject.nestedNode
        assert n1.subject not in n2.subject.nestedNode


class NodeComponentGroupTestCase(TestCase):
    def test_grouping(self):
        """Test component within node composition
        """
        n = self.create(NodeItem, UML.Node)
        c = self.create(ComponentItem, UML.Component)

        self.group(n, c)

        assert 1 == len(n.subject.ownedAttribute)
        assert 1 == len(n.subject.ownedConnector)
        assert 1 == len(c.subject.ownedAttribute)
        assert 2 == len(self.kindof(UML.ConnectorEnd))

        a1 = n.subject.ownedAttribute[0]
        a2 = c.subject.ownedAttribute[0]

        assert a1.isComposite
        assert a1 in n.subject.part

        connector = n.subject.ownedConnector[0]
        assert connector.end[0].role is a1
        assert connector.end[1].role is a2

    def test_ungrouping(self):
        """Test decomposition of component from node
        """
        n = self.create(NodeItem, UML.Node)
        c = self.create(ComponentItem, UML.Component)

        query = self.group(n, c)
        query = self.ungroup(n, c)

        assert 0 == len(n.subject.ownedAttribute)
        assert 0 == len(c.subject.ownedAttribute)
        assert 0 == len(self.kindof(UML.Property))
        assert 0 == len(self.kindof(UML.Connector))
        assert 0 == len(self.kindof(UML.ConnectorEnd))


class NodeArtifactGroupTestCase(TestCase):
    def test_grouping(self):
        """Test artifact within node deployment
        """
        n = self.create(NodeItem, UML.Node)
        a = self.create(ArtifactItem, UML.Artifact)

        self.group(n, a)

        assert 1 == len(n.subject.deployment)
        assert n.subject.deployment[0].deployedArtifact[0] is a.subject

    def test_ungrouping(self):
        """Test removal of artifact from node
        """
        n = self.create(NodeItem, UML.Node)
        a = self.create(ArtifactItem, UML.Artifact)

        query = self.group(n, a)
        query = self.ungroup(n, a)

        assert 0 == len(n.subject.deployment)
        assert 0 == len(self.kindof(UML.Deployment))


class SubsystemUseCaseGroupTestCase(TestCase):
    def test_grouping(self):
        """Test adding an use case to a subsystem
        """
        s = self.create(ComponentItem, UML.Component)
        uc1 = self.create(UseCaseItem, UML.UseCase)
        uc2 = self.create(UseCaseItem, UML.UseCase)

        self.group(s, uc1)
        assert 1 == len(uc1.subject.subject)
        self.group(s, uc2)
        assert 1 == len(uc2.subject.subject)

        # Classifier.useCase is not navigable to UseCase
        # self.assertEqual(2, len(s.subject.useCase))

    def test_grouping_with_namespace(self):
        """Test adding an use case to a subsystem (with namespace)
        """
        s = self.create(ComponentItem, UML.Component)
        uc = self.create(UseCaseItem, UML.UseCase)

        # manipulate namespace
        c = self.element_factory.create(UML.Class)
        attribute = self.element_factory.create(UML.Property)
        c.ownedAttribute = attribute

        self.group(s, uc)
        assert 1 == len(uc.subject.subject)
        assert s.subject.namespace is not uc.subject

    def test_ungrouping(self):
        """Test removal of use case from subsystem
        """
        s = self.create(ComponentItem, UML.Component)
        uc1 = self.create(UseCaseItem, UML.UseCase)
        uc2 = self.create(UseCaseItem, UML.UseCase)

        self.group(s, uc1)
        self.group(s, uc2)

        self.ungroup(s, uc1)
        assert 0 == len(uc1.subject.subject)
        # Classifier.useCase is not navigable to UseCase
        # self.assertEqual(1, len(s.subject.useCase))

        self.ungroup(s, uc2)
        assert 0 == len(uc2.subject.subject)
        # Classifier.useCase is not navigable to UseCase
        # self.assertEqual(0, len(s.subject.useCase))



"""
CommentLine -- A line that connects a comment to another model element.

"""

from gaphor.diagram.presentation import LinePresentation
from gaphor.diagram.connectors import IConnect


class CommentLineItem(LinePresentation):
    def __init__(self, id=None, model=None):
        super().__init__(id, model, style={"dash-style": (7.0, 5.0)})

    def unlink(self):
        canvas = self.canvas
        c1 = canvas.get_connection(self.head)
        c2 = canvas.get_connection(self.tail)
        if c1 and c2:
            adapter = IConnect(c1.connected, self)
            adapter.disconnect(self.head)
        super().unlink()



from gi.repository import Gtk

from gaphor.core import _, transactional
from gaphor.diagram.propertypages import PropertyPages, NamedItemPropertyPage
from gaphor.diagram.components import ComponentItem


@PropertyPages.register(ComponentItem)
class ComponentPropertyPage(NamedItemPropertyPage):
    """
    """

    def construct(self):
        page = super().construct()

        subject = self.subject

        if not subject:
            return page

        hbox = Gtk.HBox()
        page.pack_start(hbox, False, True, 0)

        button = Gtk.CheckButton(_("Indirectly instantiated"))
        button.set_active(subject.isIndirectlyInstantiated)
        button.connect("toggled", self._on_ii_change)
        hbox.pack_start(button, False, True, 0)

        return page

    def update(self):
        pass

    @transactional
    def _on_ii_change(self, button):
        """
        Called when user clicks "Indirectly instantiated" check button.
        """
        subject = self.subject
        if subject:
            subject.isIndirectlyInstantiated = button.get_active()



# vim: sw=4
"""This module contains a model element Diagram which is the abstract
representation of a UML diagram. Diagrams can be visualized and edited.

The DiagramCanvas class extends the gaphas.Canvas class.
"""

import uuid

import gaphas

from gaphor.UML.properties import umlproperty
from gaphor.UML.uml2 import Namespace, PackageableElement


class DiagramCanvas(gaphas.Canvas):
    """DiagramCanvas extends the gaphas.Canvas class.  Updates to the canvas
    can be blocked by setting the block_updates property to true.  A save
    function can be applied to all root canvas items.  Canvas items can be
    selected with an optional expression filter."""

    def __init__(self, diagram):
        """Initialize the diagram canvas with the supplied diagram.  By default,
        updates are not blocked."""

        super().__init__()
        self._diagram = diagram
        self._block_updates = False

    diagram = property(lambda s: s._diagram)

    def _set_block_updates(self, block):
        """Sets the block_updates property.  If false, the diagram canvas is
        updated immediately."""

        self._block_updates = block
        if not block:
            self.update_now()

    block_updates = property(lambda s: s._block_updates, _set_block_updates)

    def update_now(self):
        """Update the diagram canvas, unless block_updates is true."""

        if self._block_updates:
            return
        super().update_now()

    def save(self, save_func):
        """Apply the supplied save function to all root diagram items."""

        for item in self.get_root_items():
            save_func(None, item)

    def postload(self):
        """Called after the diagram canvas has loaded.  Currently does nothing.
        """

    def select(self, expression=lambda e: True):
        """Return a list of all canvas items that match expression."""

        return list(filter(expression, self.get_all_items()))


class Diagram(Namespace, PackageableElement):
    """Diagrams may contain model elements and can be owned by a Package.
    """

    def __init__(self, id=None, model=None):
        """Initialize the diagram with an optional id and element model.
        The diagram also has a canvas."""

        super().__init__(id, model)
        self.canvas = DiagramCanvas(self)

    package: umlproperty[Namespace, Namespace]

    def save(self, save_func):
        """Apply the supplied save function to this diagram and the canvas."""

        super().save(save_func)
        save_func("canvas", self.canvas)

    def postload(self):
        """Handle post-load functionality for the diagram canvas."""
        super().postload()
        self.canvas.postload()

    def create(self, type, parent=None, subject=None):
        """Create a new canvas item on the canvas. It is created with
        a unique ID and it is attached to the diagram's root item.  The type
        parameter is the element class to create.  The new element also has an
        optional parent and subject."""

        return self.create_as(type, str(uuid.uuid1()), parent, subject)

    def create_as(self, type, id, parent=None, subject=None):
        assert issubclass(type, gaphas.Item)
        item = type(id, self.model)
        if subject:
            item.subject = subject
        self.canvas.add(item, parent)
        return item

    def unlink(self):
        """Unlink all canvas items then unlink this diagram."""

        for item in self.canvas.get_all_items():
            try:
                item.unlink()
            except:
                pass

        super().unlink()



import pytest
from typing import Dict

import gaphas, gaphas.examples
from gaphor.plugins.diagramexport import DiagramExport
import gaphor.ui.menufragment


@pytest.fixture
def diagram_export():
    properties: Dict[str, str] = {}
    export_menu = gaphor.ui.menufragment.MenuFragment()

    return DiagramExport(diagrams=None, properties=properties, export_menu=export_menu)


@pytest.fixture
def canvas():
    c = gaphas.Canvas()
    c.add(gaphas.examples.Box())
    return c


def test_export_to_svg(diagram_export, canvas, tmp_path):
    f = tmp_path / "test.svg"

    diagram_export.save_svg(f, canvas)
    content = f.read_text()

    assert "<svg" in content


def test_export_to_png(diagram_export, canvas, tmp_path):
    f = tmp_path / "test.png"

    diagram_export.save_png(f, canvas)
    content = f.read_bytes()

    assert b"PNG" in content


def test_export_to_pdf(diagram_export, canvas, tmp_path):
    f = tmp_path / "test.pdf"

    diagram_export.save_pdf(f, canvas)
    content = f.read_bytes()

    assert b"%PDF" in content



"""Unit tests for transactions in Gaphor."""

from unittest import TestCase

from gaphor.core import event_handler
from gaphor.transaction import Transaction, TransactionError
from gaphor.event import TransactionBegin, TransactionCommit, TransactionRollback
from gaphor.services.eventmanager import EventManager


begins = []
commits = []
rollbacks = []


@event_handler(TransactionBegin)
def handle_begins(event):
    """Store TransactionBegin events in begins."""
    begins.append(event)


@event_handler(TransactionCommit)
def handle_commits(event):
    """Store TransactionCommit events in commits."""
    commits.append(event)


@event_handler(TransactionRollback)
def handle_rollback(event):
    """Store TransactionRollback events in rollbacks."""
    rollbacks.append(event)


class TransactionTestCase(TestCase):
    """Test case for transactions with the component registry
    enabled."""

    def setUp(self):
        """Initialize Gaphor services and register transaction event
        handlers."""

        self.event_manager = EventManager()

        self.event_manager.subscribe(handle_begins)
        self.event_manager.subscribe(handle_commits)
        self.event_manager.subscribe(handle_rollback)

        del begins[:]
        del commits[:]
        del rollbacks[:]

    def tearDown(self):
        """Finished with the test case.  Unregister event handlers that
        store transaction events."""

        self.event_manager.unsubscribe(handle_begins)
        self.event_manager.unsubscribe(handle_commits)
        self.event_manager.unsubscribe(handle_rollback)

    def test_transaction_commit(self):
        """Test committing a transaction."""

        tx = Transaction(self.event_manager)

        assert tx._stack, "Transaction has no stack"
        assert 1 == len(begins), "Incorrect number of TrasactionBegin events"
        assert 0 == len(commits), "Incorrect number of TransactionCommit events"
        assert 0 == len(rollbacks), "Incorrect number of TransactionRollback events"

        tx.commit()

        assert 1 == len(begins), "Incorrect number of TrasactionBegin events"
        assert 1 == len(commits), "Incorrect number of TransactionCommit events"
        assert 0 == len(rollbacks), "Incorrect number of TransactionRollback events"
        assert not tx._stack, "Transaction stack is not empty"

        try:
            tx.commit()
        except TransactionError:
            pass
        else:
            self.fail("Commit should not have succeeded")

    def test_transaction_rollback(self):
        """Test rolling back a transaction."""

        tx = Transaction(self.event_manager)

        assert tx._stack, "Transaction has no stack"
        assert 1 == len(begins), "Incorrect number of TrasactionBegin events"
        assert 0 == len(commits), "Incorrect number of TransactionCommit events"
        assert 0 == len(rollbacks), "Incorrect number of TransactionRollback events"

        tx.rollback()

        assert 1 == len(begins), "Incorrect number of TrasactionBegin events"
        assert 0 == len(commits), "Incorrect number of TransactionCommit events"
        assert 1 == len(rollbacks), "Incorrect number of TransactionRollback events"

        assert not tx._stack, "Transaction stack is not empty"

    def test_transaction_commit_after_rollback(self):
        """Test committing one transaction after rolling back another
        transaction."""

        tx1 = Transaction(self.event_manager)
        tx2 = Transaction(self.event_manager)

        tx2.rollback()
        tx1.commit()

        assert 1 == len(begins), "Incorrect number of TrasactionBegin events"
        assert 0 == len(commits), "Incorrect number of TransactionCommit events"
        assert 1 == len(rollbacks), "Incorrect number of TransactionRollback events"

    def test_transaction_stack(self):
        """Test the transaction stack."""

        tx1 = Transaction(self.event_manager)
        tx2 = Transaction(self.event_manager)

        try:
            self.assertRaises(TransactionError, tx1.commit)
        finally:
            tx2.rollback()
            tx1.rollback()

    def test_transaction_context(self):
        """Test the transaction context manager."""

        with Transaction(self.event_manager) as tx:

            assert isinstance(tx, Transaction), "Context is not a Transaction instance"
            assert (
                Transaction._stack
            ), "Transaction instance has no stack inside a context"

        assert not Transaction._stack, "Transaction stack should be empty"

    def test_transaction_context_error(self):
        """Test the transaction context manager with errors."""

        try:
            with Transaction(self.event_manager):
                raise TypeError("transaction error")
        except TypeError as e:
            assert "transaction error" == str(
                e
            ), "Transaction context manager did no raise correct exception"
        else:
            self.fail(
                "Transaction context manager did not raise exception when it should have"
            )



"""Support for actions in generic files.

"""

from typing import Optional, Sequence, get_type_hints
import platform


_primary = "⌘" if platform.system() == "Darwin" else "Ctrl"


def primary():
    global _primary
    return _primary


class action:
    """
    Decorator. Turns a regular function (/method) into a full blown
    Action class.

    >>> class A:
    ...     @action(name="my_action", label="my action")
    ...     def myaction(self):
    ...         print('action called')
    >>> a = A()
    >>> a.myaction()
    action called
    >>> is_action(a.myaction)
    True
    >>> for method in dir(A):
    ...     if is_action(getattr(A, method, None)):
    ...         print(method)
    myaction
    >>> A.myaction.__action__.name
    'my_action'
    >>> A.myaction.__action__.label
    'my action'
    """

    def __init__(
        self, name, label=None, tooltip=None, icon_name=None, shortcut=None, state=None
    ):
        self.scope, self.name = name.split(".", 2) if "." in name else ("win", name)
        self.label = label
        self.tooltip = tooltip
        self.icon_name = icon_name
        self.shortcut = shortcut
        self.state = state
        self.arg_type = None

    def __call__(self, func):
        type_hints = get_type_hints(func)
        if len(type_hints) == 1:
            # assume the first argument (exclusing self) is our parameter
            self.arg_type = next(iter(type_hints.values()))
        func.__action__ = self
        return func


def is_action(func):
    return hasattr(func, "__action__")


if __name__ == "__main__":
    import doctest

    doctest.testmod()



from gaphor.diagram.editors import Editor, AbstractEditor, NamedItemEditor
from gaphor.diagram.actions.action import ActionItem
from gaphor.diagram.actions.activitynodes import ForkNodeItem


Editor.register(ActionItem, NamedItemEditor)


@Editor.register(ForkNodeItem)
class ForkNodeItemEditor(AbstractEditor):
    """Text edit support for fork node join specification."""

    def __init__(self, item):
        self._item = item

    def is_editable(self, x, y):
        return True

    def get_text(self):
        """
        Get join specification text.
        """
        if self._item.subject.joinSpec:
            return self._item.subject.joinSpec
        else:
            return ""

    def update_text(self, text):
        """
        Set join specification text.
        """
        spec = self._item.subject.joinSpec
        if not spec:
            spec = text

    def key_pressed(self, pos, key):
        pass



import os
from gi.repository import GLib, Gtk

from gaphor.abc import Service
from gaphor.core import event_handler
from gaphor.ui import APPLICATION_ID
from gaphor.ui.event import FileLoaded, FileSaved


class RecentFiles(Service):
    def __init__(self, event_manager, recent_manager=None):
        self.event_manager = event_manager
        self.recent_manager = recent_manager or Gtk.RecentManager.get_default()

        self.event_manager.subscribe(self._on_filename_changed)

    def shutdown(self):
        self.event_manager.unsubscribe(self._on_filename_changed)

    @event_handler(FileLoaded, FileSaved)
    def _on_filename_changed(self, event):
        filename = event.filename
        if not filename:
            return
        uri = GLib.filename_to_uri(os.path.abspath(filename))
        meta = Gtk.RecentData()
        meta.app_name = APPLICATION_ID
        meta.app_exec = f"{GLib.get_prgname()} %u"
        meta.mime_type = "application/x-gaphor"
        self.recent_manager.add_full(uri, meta)



from gaphor.diagram.components.artifact import ArtifactItem
from gaphor.diagram.components.component import ComponentItem
from gaphor.diagram.components.connector import ConnectorItem
from gaphor.diagram.components.node import NodeItem


def _load():
    from gaphor.diagram.components import (
        componentsgrouping,
        connectorconnect,
        componentspropertypage,
    )


_load()



"""
Control flow and object flow implementation.

Contains also implementation to split flows using activity edge connectors.
"""


from gaphor import UML
from gaphor.UML.modelfactory import stereotypes_str
from gaphor.diagram.presentation import LinePresentation
from gaphor.diagram.support import represents
from gaphor.diagram.shapes import Box, Text, EditableText, draw_arrow_tail


@represents(UML.ControlFlow)
class FlowItem(LinePresentation):
    """
    Representation of control flow and object flow. Flow item has name and
    guard. It can be splitted into two flows with activity edge connectors.
    """

    def __init__(self, id=None, model=None):
        super().__init__(id, model)

        self.shape_tail = Box(
            Text(
                text=lambda: stereotypes_str(self.subject),
                style={"min-width": 0, "min-height": 0},
            ),
            EditableText(text=lambda: self.subject.name or ""),
        )

        self.watch("subject[NamedElement].name")
        self.watch("subject.appliedStereotype.classifier.name")

        self.shape_middle = EditableText(
            text=lambda: self.subject and self.subject.guard or ""
        )

        self.watch("subject[ControlFlow].guard")
        self.watch("subject[ObjectFlow].guard")

        self.draw_tail = draw_arrow_tail



"""Internationalization (i18n) support for Gaphor.

Here the _() function is defined that is used to translate text into
your native language."""

__all__ = ["_"]

import os

import gettext
import importlib_metadata

localedir = os.path.join(
    importlib_metadata.distribution("gaphor").locate_file("gaphor/data/locale")
)

try:

    catalog = gettext.Catalog("gaphor", localedir=localedir)
    _ = catalog.gettext

except OSError as e:

    def _(s):
        return s



from gaphor import UML
from gi.repository import Gtk

from gaphor.tests.testcase import TestCase
from gaphor.ui.diagrampage import DiagramPage
from gaphor.ui.diagramtoolbox import TOOLBOX_ACTIONS


class WindowOwner:
    """
    Placeholder object for a MainWindow. Should provide just enough
    methods to make the tests work.
    """


class DiagramToolboxTestCase(TestCase):

    services = [
        "event_manager",
        "component_registry",
        "element_factory",
        "properties",
        "main_window",
        "export_menu",
        "tools_menu",
    ]

    def setUp(self):
        TestCase.setUp(self)

        self.tab = DiagramPage(
            WindowOwner(),
            self.get_service("event_manager"),
            self.element_factory,
            self.get_service("properties"),
        )
        self.tab.diagram = self.diagram

        self.window = Gtk.Window.new(Gtk.WindowType.TOPLEVEL)
        self.window.add(self.tab.construct())
        self.window.show()

    def tearDown(self):
        self.window.destroy()
        TestCase.tearDown(self)

    def test_toolbox_actions_shortcut_unique(self):

        shortcuts = {}

        for category, items in TOOLBOX_ACTIONS:
            for action_name, label, icon_name, shortcut in items:
                try:
                    shortcuts[shortcut].append(action_name)
                except KeyError:
                    shortcuts[shortcut] = [action_name]

        for key, val in list(shortcuts.items()):
            if key is not None:
                assert len(val) == 1, "Duplicate toolbox shortcut"

    def test_standalone_construct_with_diagram(self):
        pass  # is setUp()

    def _test_placement_action(self, action):
        tool = self.tab.toolbox.get_tool(action)

        # Ensure the factory is working
        tool.create_item((0, 0))
        self.diagram.canvas.update()

    def test_placement_pointer(self):
        tool = self.tab.toolbox.get_tool("toolbox-pointer")

        assert tool

    def test_placement_comment(self):
        self._test_placement_action("toolbox-comment")

    def test_placement_comment_line(self):
        self._test_placement_action("toolbox-comment-line")

    # Classes:

    def test_placement_class(self):
        self._test_placement_action("toolbox-class")

    def test_placement_interface(self):
        self._test_placement_action("toolbox-interface")

    def test_placement_package(self):
        self._test_placement_action("toolbox-package")

    def test_placement_association(self):
        self._test_placement_action("toolbox-association")

    def test_placement_dependency(self):
        self._test_placement_action("toolbox-dependency")

    def test_placement_generalization(self):
        self._test_placement_action("toolbox-generalization")

    def test_placement_implementation(self):
        self._test_placement_action("toolbox-implementation")

    # Components:

    def test_placement_component(self):
        self._test_placement_action("toolbox-component")

    def test_placement_node(self):
        self._test_placement_action("toolbox-node")

    def test_placement_artifact(self):
        self._test_placement_action("toolbox-artifact")

    # Actions:

    def test_placement_action(self):
        self._test_placement_action("toolbox-action")

    def test_placement_initial_node(self):
        self._test_placement_action("toolbox-initial-node")

    def test_placement_activity_final_node(self):
        self._test_placement_action("toolbox-activity-final-node")

    def test_placement_flow_final_node(self):
        self._test_placement_action("toolbox-flow-final-node")

    def test_placement_decision_node(self):
        self._test_placement_action("toolbox-decision-node")

    def test_placement_fork_node(self):
        self._test_placement_action("toolbox-fork-node")

    def test_placement_object_node(self):
        self._test_placement_action("toolbox-object-node")
        assert 1 == len(self.kindof(UML.ObjectNode))

    def test_placement_partition(self):
        self._test_placement_action("toolbox-partition")
        assert 0 == len(self.kindof(UML.ActivityPartition))

    def test_placement_flow(self):
        self._test_placement_action("toolbox-flow")

    # Use cases:

    def test_use_case(self):
        self._test_placement_action("toolbox-use-case")

    def test_actor(self):
        self._test_placement_action("toolbox-actor")

    def test_use_case_association(self):
        self._test_placement_action("toolbox-use-case-association")

    def test_include(self):
        self._test_placement_action("toolbox-include")

    def test_extend(self):
        self._test_placement_action("toolbox-extend")

    # Profiles:

    def test_profile(self):
        self._test_placement_action("toolbox-profile")

    def test_metaclass(self):
        self._test_placement_action("toolbox-metaclass")

    def test_stereotype(self):
        self._test_placement_action("toolbox-stereotype")

    def test_extension(self):
        self._test_placement_action("toolbox-extension")



"""
Support classes for dealing with text.
"""

from typing import Any, Tuple, Dict, TypeVar

from enum import Enum
import cairo
from gi.repository import GLib, Pango, PangoCairo

from gaphas.freehand import FreeHandCairoContext
from gaphas.painter import CairoBoundingBoxContext

Font = TypeVar("Font", Dict[str, Any], str)


class TextAlign(Enum):
    LEFT = "left"
    CENTER = "center"
    RIGHT = "right"


class VerticalAlign(Enum):
    TOP = "top"
    MIDDLE = "middle"
    BOTTOM = "bottom"


class FontStyle(Enum):
    NORMAL = "normal"
    ITALIC = "italic"


class FontWeight(Enum):
    NORMAL = "normal"
    BOLD = "bold"


class TextDecoration(Enum):
    NONE = "none"
    UNDERLINE = "underline"


def text_draw_focus_box(context, x, y, w, h):
    if context.hovered or context.focused or context.draw_all:
        cr = context.cairo
        cr.save()
        try:
            cr.set_source_rgb(0.6, 0.6, 0.6)
            cr.set_line_width(0.5)
            cr.rectangle(x, y, w, h)
            cr.stroke()
        finally:
            cr.restore()


def text_size(
    cr, text: str, font: Font, width=-1, default_size: Tuple[int, int] = (0, 0)
) -> Tuple[int, int]:
    if not text:
        return default_size

    layout = _text_layout(cr, text, font, width)
    return layout.get_pixel_size()  # type: ignore


def text_draw(cr, text, font, calculate_pos, width=-1, default_size=(0, 0)):
    """
    Draw text relative to (x, y).
    text - text to print (utf8)
    font - The font to render in, either a string, or a dict with "font", and optional "font-style" and "text-decoration"
    calculate_pos - callback for text positioning
    default_size - default text size, when no text is provided
    """

    if text:
        layout = _text_layout(cr, text, font, width)
        w, h = layout.get_pixel_size()
    else:
        layout = None
        w, h = default_size

    x, y = calculate_pos(w, h)

    cr.move_to(x, y)

    if layout:
        _pango_cairo_show_layout(cr, layout)

    return (x, y, w, h)


def _text_layout(cr, text, font, width):
    underline = False
    layout = _pango_cairo_create_layout(cr)

    if isinstance(font, dict):
        fd = Pango.FontDescription.from_string(font["font"])
        font_weight = font.get("font-weight")
        font_style = font.get("font-style")
        if font_weight:
            fd.set_weight(getattr(Pango.Weight, font_weight.name))
        if font_style:
            fd.set_style(getattr(Pango.Style, font_style.name))
        layout.set_font_description(fd)
        underline = (
            font.get("text-decoration", TextDecoration.NONE) == TextDecoration.UNDERLINE
        )
    else:
        layout.set_font_description(Pango.FontDescription.from_string(font))

    if underline:
        # TODO: can this be done via Pango attributes instead?
        layout.set_markup(f"<u>{GLib.markup_escape_text(text)}</u>", length=-1)
    else:
        layout.set_text(text, length=-1)
    layout.set_width(int(width * Pango.SCALE))
    return layout


def _pango_cairo_create_layout(cr):
    """
    Deal with different types of contexts that are passed down,
    namely FreeHandCairoContext and CairoBoundingBoxContext.
    PangoCairo expects a true cairo.Context.
    """
    if isinstance(cr, CairoBoundingBoxContext):
        cr = cr._cairo
    if isinstance(cr, FreeHandCairoContext):
        cr = cr.cr

    assert isinstance(cr, cairo.Context), f"cr should be a true Cairo.Context, not {cr}"

    return PangoCairo.create_layout(cr)


def _pango_cairo_show_layout(cr, layout):
    if isinstance(cr, FreeHandCairoContext):
        PangoCairo.show_layout(cr.cr, layout)
    elif isinstance(cr, CairoBoundingBoxContext):
        w, h = layout.get_pixel_size()
        cr.rel_line_to(w, h)
        cr.stroke()
    else:
        PangoCairo.show_layout(cr, layout)


def text_point_in_box(bounding_box, text_size, text_align, vertical_align):
    x, y, width, height = bounding_box
    w, h = text_size

    if text_align is TextAlign.CENTER:
        x += (width - w) / 2
    elif text_align is TextAlign.RIGHT:
        x += width - w

    if vertical_align is VerticalAlign.MIDDLE:
        y += (height - h) / 2
    elif vertical_align is VerticalAlign.BOTTOM:
        y += height

    return x, y


def text_point_at_line(points, size, text_align):
    """
    Provide a position (x, y) to draw a text close to a line.

    Parameters:
     - points:  the line points, a list of (x, y) points
     - size:    size of the text, a (width, height) tuple
     - text_align: alignment to the line: left, beginning of the line, center, middle and right: end of the line
    """

    if text_align == TextAlign.LEFT:
        p0 = points[0]
        p1 = points[1]
        x, y = _text_point_at_line_end(size, p0, p1)
    elif text_align == TextAlign.CENTER:
        p0, p1 = middle_segment(points)
        x, y = _text_point_at_line_center(size, p0, p1)
    elif text_align == TextAlign.RIGHT:
        p0 = points[-1]
        p1 = points[-2]
        x, y = _text_point_at_line_end(size, p0, p1)

    return x, y


def middle_segment(points):
    """
    Get middle line segment.
    """
    m = len(points) // 2
    assert m - 1 >= 0 and m < len(points)
    return points[m - 1], points[m]


def _text_point_at_line_end(size, p1, p2):
    """
    Calculate position of the text relative to a line defined by points
    p1 and p2.

    Parameters:
     - size: text size, a (width, height) tuple
     - p1:      beginning of line segment
     - p2:      end of line segment
    """
    name_dx = 0.0
    name_dy = 0.0
    ofs = 5

    dx = float(p2[0]) - float(p1[0])
    dy = float(p2[1]) - float(p1[1])

    name_w, name_h = size

    if dy == 0:
        rc = 1000.0  # quite a lot...
    else:
        rc = dx / dy
    abs_rc = abs(rc)
    h = dx > 0  # right side of the box
    v = dy > 0  # bottom side

    if abs_rc > 6:
        # horizontal line
        if h:
            name_dx = ofs
            name_dy = -ofs - name_h
        else:
            name_dx = -ofs - name_w
            name_dy = -ofs - name_h
    elif 0 <= abs_rc <= 0.2:
        # vertical line
        if v:
            name_dx = -ofs - name_w
            name_dy = ofs
        else:
            name_dx = -ofs - name_w
            name_dy = -ofs - name_h
    else:
        # Should both items be placed on the same side of the line?
        r = abs_rc < 1.0

        # Find out alignment of text (depends on the direction of the line)
        align_left = h ^ r
        align_bottom = v ^ r
        if align_left:
            name_dx = ofs
        else:
            name_dx = -ofs - name_w
        if align_bottom:
            name_dy = -ofs - name_h
        else:
            name_dy = ofs
    return p1[0] + name_dx, p1[1] + name_dy


# hint tuples to move text depending on quadrant
WIDTH_HINT = (-1, -1, 0)
PADDING_HINT = (1, 1, -1)
EPSILON = 1e-6


def _text_point_at_line_center(size, p1, p2):
    """
    Calculate position of the text relative to a line defined by points
    p1 and p2.

    Parameters:
     - size:    text size, a (width, height) tuple
     - p1:      beginning of line
     - p2:      end of line
    """
    x0 = (p1[0] + p2[0]) / 2.0
    y0 = (p1[1] + p2[1]) / 2.0
    dx = p2[0] - p1[0]
    dy = p2[1] - p1[1]
    ofs = 3

    if abs(dx) < EPSILON:
        d1 = -1.0
        d2 = 1.0
    elif abs(dy) < EPSILON:
        d1 = 0.0
        d2 = 0.0
    else:
        d1 = dy / dx
        d2 = abs(d1)

    width, height = size

    # move to center and move by delta depending on line angle
    if d2 < 0.5774:  # <0, 30>, <150, 180>, <-180, -150>, <-30, 0>
        # horizontal mode
        w2 = width / 2.0
        hint = w2 * d2

        x = x0 - w2
        y = y0 + hint + ofs
    else:
        # much better in case of vertical lines

        # determine quadrant, we are interested in 1 or 3 and 2 or 4
        # see hint tuples below
        h2 = height / 2.0
        q = (d1 > 0) - (d1 < 0)
        if abs(dx) < EPSILON:
            hint = 0
        else:
            hint = h2 / d2

        x = x0 - hint + width * WIDTH_HINT[q]
        x = x0 - (ofs + hint) * PADDING_HINT[q] + width * WIDTH_HINT[q]
        y = y0 - h2

    return x, y



"""
Formatting of UML model elements into text tests.
"""

import pytest

from gaphor.services.eventmanager import EventManager
from gaphor.UML.elementfactory import ElementFactory
from gaphor.UML.umlfmt import format
import gaphor.UML.uml2 as UML


@pytest.fixture
def factory():
    event_manager = EventManager()
    return ElementFactory(event_manager)


def test_simple_format(factory):
    """Test simple attribute formatting
    """
    a = factory.create(UML.Property)
    a.name = "myattr"
    assert "+ myattr" == format(a)

    a.typeValue = "int"
    assert "+ myattr: int" == format(a)



"""A generic way to handle errors in GUI applications.

This module also contains a ErrorHandlerAspect, which can be easily attached
to a class' method and will raise the error dialog when the method exits with
an exception.
"""

from gi.repository import Gtk
import sys
import pdb

from gaphor.i18n import _


def error_handler(message=None, exc_info=None):
    exc_type, exc_value, exc_traceback = exc_info or sys.exc_info()

    if not exc_type:
        return

    if not message:
        message = _("An error occurred.")

    buttons = Gtk.ButtonsType.OK
    message = f"{message}\n\nTechnical details:\n\t{exc_type}\n\t{exc_value}"

    if __debug__ and sys.stdin.isatty():
        buttons = Gtk.ButtonsType.YES_NO
        message += _(
            "\n\nDo you want to debug?\n(Gaphor should have been started from the command line)"
        )

    dialog = Gtk.MessageDialog(
        None,
        Gtk.DialogFlags.MODAL | Gtk.DialogFlags.DESTROY_WITH_PARENT,
        Gtk.MessageType.ERROR,
        buttons,
        message,
    )
    answer = dialog.run()
    dialog.destroy()
    if answer == Gtk.ResponseType.YES:
        pdb.post_mortem(exc_traceback)



"""Service dedicated to exporting diagrams to a variety of file formats."""

import os
import logging

import cairo
from gaphas.freehand import FreeHandPainter
from gaphas.painter import ItemPainter, BoundingBoxPainter
from gaphas.view import Context, View

from gaphor.core import _, action
from gaphor.abc import Service, ActionProvider
from gaphor.ui.filedialog import FileDialog
from gaphor.ui.questiondialog import QuestionDialog

logger = logging.getLogger(__name__)


def paint(view, cr):
    view.painter.paint(Context(cairo=cr, items=view.canvas.get_all_items(), area=None))


class DiagramExport(Service, ActionProvider):
    """
    Service for exporting diagrams as images (SVG, PNG, PDF).
    """

    def __init__(self, diagrams, properties, export_menu):
        self.diagrams = diagrams
        self.properties = properties
        self.export_menu = export_menu
        export_menu.add_actions(self)

    def shutdown(self):
        self.export_menu.remove_actions(self)

    def save_dialog(self, diagram, title, ext):

        filename = (diagram.name or "export") + ext
        file_dialog = FileDialog(title, action="save", filename=filename)

        save = False
        while True:
            filename = file_dialog.selection
            if os.path.exists(filename):
                question = (
                    _(
                        "The file %s already exists. Do you want to "
                        "replace it with the file you are exporting "
                        "to?"
                    )
                    % filename
                )
                question_dialog = QuestionDialog(question)
                answer = question_dialog.answer
                question_dialog.destroy()
                if answer:
                    save = True
                    break
            else:
                save = True
                break

        file_dialog.destroy()

        if save and filename:
            return filename

    def update_painters(self, view):
        sloppiness = self.properties.get("diagram.sloppiness", 0)

        if sloppiness:
            view.painter = FreeHandPainter(ItemPainter(), sloppiness)
            view.bounding_box_painter = FreeHandPainter(
                BoundingBoxPainter(), sloppiness
            )
        else:
            view.painter = ItemPainter()

    def render(self, canvas, new_surface):
        view = View(canvas)

        self.update_painters(view)

        # Update bounding boxes with a temporary CairoContext
        # (used for stuff like calculating font metrics)
        tmpsurface = cairo.ImageSurface(cairo.FORMAT_ARGB32, 0, 0)
        tmpcr = cairo.Context(tmpsurface)
        view.update_bounding_box(tmpcr)
        tmpcr.show_page()
        tmpsurface.flush()

        w, h = view.bounding_box.width, view.bounding_box.height
        surface = new_surface(w, h)
        cr = cairo.Context(surface)
        view.matrix.translate(-view.bounding_box.x, -view.bounding_box.y)
        paint(view, cr)
        cr.show_page()
        return surface

    def save_svg(self, filename, canvas):
        surface = self.render(canvas, lambda w, h: cairo.SVGSurface(filename, w, h))
        surface.flush()
        surface.finish()

    def save_png(self, filename, canvas):
        surface = self.render(
            canvas,
            lambda w, h: cairo.ImageSurface(
                cairo.FORMAT_ARGB32, int(w + 1), int(h + 1)
            ),
        )
        surface.write_to_png(filename)

    def save_pdf(self, filename, canvas):
        surface = self.render(canvas, lambda w, h: cairo.PDFSurface(filename, w, h))
        surface.flush()
        surface.finish()

    @action(
        name="file-export-svg",
        label=_("Export to SVG"),
        tooltip=_("Export the diagram to SVG"),
    )
    def save_svg_action(self):
        title = _("Export diagram to SVG")
        ext = ".svg"
        diagram = self.diagrams.get_current_diagram()
        filename = self.save_dialog(diagram, title, ext)
        if filename:
            self.save_svg(filename, diagram.canvas)

    @action(
        name="file-export-png",
        label=_("Export to PNG"),
        tooltip=_("Export the diagram to PNG"),
    )
    def save_png_action(self):
        title = _("Export diagram to PNG")
        ext = ".png"
        diagram = self.diagrams.get_current_diagram()
        filename = self.save_dialog(diagram, title, ext)
        if filename:
            self.save_png(filename, diagram.canvas)

    @action(
        name="file-export-pdf",
        label=_("Export to PDF"),
        tooltip=_("Export the diagram to PDF"),
    )
    def save_pdf_action(self):
        title = _("Export diagram to PDF")
        ext = ".pdf"
        diagram = self.diagrams.get_current_diagram()
        filename = self.save_dialog(diagram, title, ext)
        if filename:
            self.save_pdf(filename, diagram.canvas)



import pytest
from gaphor import UML
from gaphor.UML.elementfactory import ElementFactory
from gaphor.plugins.xmiexport.exportmodel import XMIExport


@pytest.fixture
def element_factory():
    ef = ElementFactory()
    p = ef.create(UML.Package).name = "package"
    c1 = ef.create(UML.Class)
    c1.name = "class"
    c2 = ef.create(UML.Class)
    c2.name = "class"
    i1 = ef.create(UML.Interface)
    i1.name = "interface"
    c1.ownedAttribute = ef.create(UML.Property)
    c1.ownedAttribute[0].name = "attr"
    c1.ownedOperation = ef.create(UML.Operation)
    c1.ownedOperation[0].name = "oper"
    c1.ownedOperation[0].formalParameter = ef.create(UML.Parameter)

    UML.model.create_dependency(c1, c2)
    UML.model.create_generalization(c1, c2)
    UML.model.create_association(c1, c2)
    return ef


def test_xmi_export(element_factory, tmp_path):
    exporter = XMIExport(element_factory)
    f = tmp_path / "test.gaphor"

    exporter.export(f)

    content = f.read_text()

    assert '<XMI xmi.version="2.1"' in content



"""
Flow item connection adapters tests.
"""

from typing import Type
from gaphor.tests import TestCase
from gaphor import UML
from gaphor.diagram.actions.action import ActionItem
from gaphor.diagram.actions.flow import FlowItem
from gaphor.diagram.actions.activitynodes import (
    InitialNodeItem,
    ForkNodeItem,
    FlowFinalNodeItem,
    ActivityFinalNodeItem,
    ActivityNodeItem,
    DecisionNodeItem,
)
from gaphor.diagram.actions.objectnode import ObjectNodeItem


class FlowItemBasicNodesConnectionTestCase(TestCase):
    """
    Tests for flow item connecting to basic activity nodes.
    """

    def test_initial_node_glue(self):
        """Test flow item gluing to initial node item."""

        flow = self.create(FlowItem)
        node = self.create(InitialNodeItem, UML.InitialNode)

        # tail may not connect to initial node item
        allowed = self.allow(flow, flow.tail, node)
        assert not allowed

        allowed = self.allow(flow, flow.head, node)
        assert allowed

    def test_flow_final_node_glue(self):
        """Test flow item gluing to flow final node item."""

        flow = self.create(FlowItem)
        node = self.create(FlowFinalNodeItem, UML.FlowFinalNode)

        # head may not connect to flow final node item
        allowed = self.allow(flow, flow.head, node)
        assert not allowed

        allowed = self.allow(flow, flow.tail, node)
        assert allowed

    def test_activity_final_node_glue(self):
        """Test flow item gluing to activity final node item
        """
        flow = self.create(FlowItem)
        node = self.create(ActivityFinalNodeItem, UML.ActivityFinalNode)

        # head may not connect to activity final node item
        glued = self.allow(flow, flow.head, node)
        assert not glued

        glued = self.allow(flow, flow.tail, node)
        assert glued


class FlowItemObjectNodeTestCase(TestCase):
    """
    Flow item connecting to object node item tests.
    """

    def test_glue_to_object_node(self):
        flow = self.create(FlowItem)
        onode = self.create(ObjectNodeItem, UML.ObjectNode)
        glued = self.allow(flow, flow.head, onode)
        assert glued

    def test_connect_to_object_node(self):
        flow = self.create(FlowItem)
        anode = self.create(ActionItem, UML.Action)
        onode = self.create(ObjectNodeItem, UML.ObjectNode)

        self.connect(flow, flow.head, anode)
        self.connect(flow, flow.tail, onode)
        assert flow.subject
        assert isinstance(flow.subject, UML.ObjectFlow)

        self.disconnect(flow, flow.head)
        self.disconnect(flow, flow.tail)

        # opposite connection
        self.connect(flow, flow.head, onode)
        self.connect(flow, flow.tail, anode)
        assert flow.subject
        assert isinstance(flow.subject, UML.ObjectFlow)

    def test_object_flow_reconnect(self):
        flow = self.create(FlowItem)
        a1 = self.create(ActionItem, UML.Action)
        o1 = self.create(ObjectNodeItem, UML.ObjectNode)
        o2 = self.create(ObjectNodeItem, UML.ObjectNode)

        # connect: a1 -> o1
        self.connect(flow, flow.head, a1)
        self.connect(flow, flow.tail, o1)

        f = flow.subject
        f.name = "tname"
        f.guard = "tguard"

        # reconnect: a1 -> o2
        self.connect(flow, flow.tail, o2)

        assert 0 == len(a1.subject.incoming)
        assert 1 == len(a1.subject.outgoing)
        # no connections to o1
        self.assertEqual(0, len(o1.subject.incoming))
        assert 0 == len(o1.subject.outgoing)
        # connections to o2 instead
        self.assertEqual(1, len(o2.subject.incoming))
        assert 0 == len(o2.subject.outgoing)

        assert 1 == len(self.kindof(UML.ObjectFlow))
        # one guard
        self.assertEqual("tname", flow.subject.name)
        assert "tguard" == flow.subject.guard

    def test_control_flow_reconnection(self):
        """Test control flow becoming object flow due to reconnection
        """
        flow = self.create(FlowItem)
        a1 = self.create(ActionItem, UML.Action)
        a2 = self.create(ActionItem, UML.Action)
        o1 = self.create(ObjectNodeItem, UML.ObjectNode)

        # connect with control flow: a1 -> a2
        self.connect(flow, flow.head, a1)
        self.connect(flow, flow.tail, a2)

        f = flow.subject
        f.name = "tname"
        f.guard = "tguard"

        # reconnect with object flow: a1 -> o1
        self.connect(flow, flow.tail, o1)

        assert 0 == len(a1.subject.incoming)
        assert 1 == len(a1.subject.outgoing)
        # no connections to a2
        self.assertEqual(0, len(a2.subject.incoming))
        assert 0 == len(a2.subject.outgoing)
        # connections to o1 instead
        self.assertEqual(1, len(o1.subject.incoming))
        assert 0 == len(o1.subject.outgoing)

        assert 0 == len(self.kindof(UML.ControlFlow))
        assert 1 == len(self.kindof(UML.ObjectFlow))
        # one guard, not changed
        self.assertEqual("tname", flow.subject.name)
        assert "tguard" == flow.subject.guard


class FlowItemActionTestCase(TestCase):
    """
    Flow item connecting to action item tests.
    """

    def test_glue(self):
        """Test flow item gluing to action items."""

        flow = self.create(FlowItem)
        a1 = self.create(ActionItem, UML.Action)
        a2 = self.create(ActionItem, UML.Action)

        glued = self.allow(flow, flow.head, a1)
        assert glued

        self.connect(flow, flow.head, a1)

        glued = self.allow(flow, flow.tail, a2)
        assert glued

    def test_connect_to_action_item(self):
        flow = self.create(FlowItem)
        a1 = self.create(ActionItem, UML.Action)
        a2 = self.create(ActionItem, UML.Action)

        self.connect(flow, flow.head, a1)
        self.connect(flow, flow.tail, a2)

        assert isinstance(flow.subject, UML.ControlFlow)

        assert 0 == len(a1.subject.incoming)
        assert 1 == len(a2.subject.incoming)
        assert 1 == len(a1.subject.outgoing)
        assert 0 == len(a2.subject.outgoing)

        assert flow.subject in a1.subject.outgoing
        assert flow.subject.source is a1.subject
        assert flow.subject in a2.subject.incoming
        assert flow.subject.target is a2.subject

    def test_disconnect_from_action_item(self):
        """Test flow item disconnection from action items
        """
        flow = self.create(FlowItem)
        a1 = self.create(ActionItem, UML.Action)
        a2 = self.create(ActionItem, UML.Action)

        self.connect(flow, flow.head, a1)
        self.connect(flow, flow.tail, a2)

        self.disconnect(flow, flow.head)
        assert flow.subject is None
        assert 0 == len(a1.subject.incoming)
        assert 0 == len(a2.subject.incoming)
        assert 0 == len(a1.subject.outgoing)
        assert 0 == len(a2.subject.outgoing)

    def test_reconnect(self):
        """Test flow item reconnection
        """
        flow = self.create(FlowItem)
        a1 = self.create(ActionItem, UML.Action)
        a2 = self.create(ActionItem, UML.Action)
        a3 = self.create(ActionItem, UML.Action)

        # a1 -> a2
        self.connect(flow, flow.head, a1)
        self.connect(flow, flow.tail, a2)
        f = flow.subject
        f.name = "tname"
        f.guard = "tguard"

        # reconnect: a1 -> a3
        self.connect(flow, flow.tail, a3)

        assert 0 == len(a1.subject.incoming)
        assert 1 == len(a1.subject.outgoing)
        # no connections to a2
        self.assertEqual(0, len(a2.subject.incoming))
        assert 0 == len(a2.subject.outgoing)
        # connections to a3 instead
        self.assertEqual(1, len(a3.subject.incoming))
        assert 0 == len(a3.subject.outgoing)

        assert 1 == len(self.kindof(UML.ControlFlow))
        # one guard
        self.assertEqual("tname", flow.subject.name)
        assert "tguard" == flow.subject.guard

    def test_object_flow_reconnection(self):
        """Test object flow becoming control flow due to reconnection
        """
        flow = self.create(FlowItem)
        a1 = self.create(ActionItem, UML.Action)
        a2 = self.create(ActionItem, UML.Action)
        o1 = self.create(ObjectNodeItem, UML.ObjectNode)

        # connect with control flow: a1 -> o1
        self.connect(flow, flow.head, a1)
        self.connect(flow, flow.tail, o1)

        f = flow.subject
        f.name = "tname"
        f.guard = "tguard"

        # reconnect with object flow: a1 -> a2
        self.connect(flow, flow.tail, a2)

        assert 0 == len(a1.subject.incoming)
        assert 1 == len(a1.subject.outgoing)
        # no connections to o1
        self.assertEqual(0, len(o1.subject.incoming))
        assert 0 == len(o1.subject.outgoing)
        # connections to a2 instead
        self.assertEqual(1, len(a2.subject.incoming))
        assert 0 == len(a2.subject.outgoing)

        assert 1 == len(self.kindof(UML.ControlFlow))
        assert 0 == len(self.kindof(UML.ObjectFlow))
        # one guard, not changed
        self.assertEqual("tname", flow.subject.name)
        assert "tguard" == flow.subject.guard


class FlowItemDesisionAndForkNodes:
    """
    Base class for flow connecting to decision and fork nodes.

    See `FlowItemDecisionNodeTestCase` and `FlowItemForkNodeTestCase` test
    cases.

    Not tested yet

    - If a join node has an incoming object flow, it must have an outgoing
      object flow, otherwise, it must have an outgoing control flow.
    - The edges coming into and out of a fork node must be either all
      object flows or all control flows.
    """

    item_cls: Type[UML.Presentation]
    fork_node_cls: Type[UML.ControlNode]
    join_node_cls: Type[UML.ControlNode]

    def test_glue(self):
        """Test decision/fork nodes glue
        """
        flow = self.create(FlowItem)
        action = self.create(ActionItem, UML.Action)
        node = self.create(self.item_cls, self.join_node_cls)

        glued = self.allow(flow, flow.head, node)
        assert glued

        self.connect(flow, flow.head, action)

        glued = self.allow(flow, flow.tail, node)
        assert glued

    def test_node_class_change(self):
        """ Test node incoming edges

        Connection scheme is presented below::

                  head  tail
            [ a1 ]--flow1-->    |
                             [ jn ] --flow3--> [ a3 ]
            [ a2 ]--flow2-->    |

        Node class changes due to two incoming edges and one outgoing edge.
        """
        flow1 = self.create(FlowItem)
        flow2 = self.create(FlowItem)
        flow3 = self.create(FlowItem)
        a1 = self.create(ActionItem, UML.Action)
        a2 = self.create(ActionItem, UML.Action)
        a3 = self.create(ActionItem, UML.Action)
        jn = self.create(self.item_cls, self.fork_node_cls)

        assert isinstance(jn.subject, self.fork_node_cls)

        # connect actions first
        self.connect(flow1, flow1.head, a1)
        self.connect(flow2, flow2.head, a2)
        self.connect(flow3, flow3.tail, a2)

        # connect to the node
        self.connect(flow1, flow1.tail, jn)
        self.connect(flow2, flow2.tail, jn)
        self.connect(flow3, flow3.head, jn)

        # node class changes
        self.assertTrue(isinstance(jn.subject, self.join_node_cls))

    def test_outgoing_edges(self):
        """Test outgoing edges


        Connection scheme is presented below::

                   head  tail    | --flow2-->[ a2 ]
            [ a1 ] --flow1--> [ jn ]
                                 | --flow3-->[ a3 ]
        """
        flow1 = self.create(FlowItem)
        flow2 = self.create(FlowItem)
        flow3 = self.create(FlowItem)
        a1 = self.create(ActionItem, UML.Action)
        a2 = self.create(ActionItem, UML.Action)
        a3 = self.create(ActionItem, UML.Action)
        jn = self.create(self.item_cls, self.join_node_cls)

        # connect actions first
        self.connect(flow1, flow1.head, a1)
        self.connect(flow2, flow2.tail, a2)
        self.connect(flow3, flow3.tail, a2)

        # connect to the node
        self.connect(flow1, flow1.tail, jn)
        assert isinstance(jn.subject, self.join_node_cls)

        self.connect(flow2, flow2.head, jn)
        assert isinstance(jn.subject, self.join_node_cls)

        assert 1 == len(jn.subject.incoming)
        assert 1 == len(jn.subject.outgoing)
        assert flow1.subject in jn.subject.incoming
        assert flow2.subject in jn.subject.outgoing

        self.connect(flow3, flow3.head, jn)
        assert 2 == len(jn.subject.outgoing)

        assert isinstance(jn.subject, self.fork_node_cls), f"{jn.subject}"

    def test_combined_nodes_connection(self):
        """Test combined nodes connection

        Connection scheme is presented below::

                   head  tail    |   --flow2--> [ a2 ]
            [ a1 ] --flow1--> [ jn ]
            [ a4 ] --flow4-->    |   --flow3--> [ a3 ]

        Flow `flow4` will force the node to become a combined node.
        """
        flow1 = self.create(FlowItem)
        flow2 = self.create(FlowItem)
        flow3 = self.create(FlowItem)
        flow4 = self.create(FlowItem)
        a1 = self.create(ActionItem, UML.Action)
        a2 = self.create(ActionItem, UML.Action)
        a3 = self.create(ActionItem, UML.Action)
        a4 = self.create(ActionItem, UML.Action)
        jn = self.create(self.item_cls, self.join_node_cls)

        # connect actions first
        self.connect(flow1, flow1.head, a1)
        self.connect(flow2, flow2.tail, a2)
        self.connect(flow3, flow3.tail, a2)
        self.connect(flow4, flow4.head, a4)

        # connect to the node
        self.connect(flow1, flow1.tail, jn)
        self.connect(flow2, flow2.head, jn)
        self.connect(flow3, flow3.head, jn)

        self.connect(flow4, flow4.tail, jn)
        assert isinstance(jn.subject, self.join_node_cls)
        assert jn.combined is not None

        # check node combination
        self.assertTrue(1, len(jn.subject.outgoing))
        assert 1, len(jn.combined.incoming)
        assert jn.subject.outgoing[0] is jn.combined.incoming[0]

    def test_combined_node_disconnection(self):
        """Test combined nodes disconnection

        Connection scheme is presented below::

                   head  tail    |   --flow2--> [ a2 ]
            [ a1 ] --flow1--> [ jn ]
            [ a4 ] --flow4-->    |   --flow3--> [ a3 ]

        Flow `flow4` will force the node to become a combined node.
        """
        canvas = self.diagram.canvas

        flow1 = self.create(FlowItem)
        flow2 = self.create(FlowItem)
        flow3 = self.create(FlowItem)
        flow4 = self.create(FlowItem)
        a1 = self.create(ActionItem, UML.Action)
        a2 = self.create(ActionItem, UML.Action)
        a3 = self.create(ActionItem, UML.Action)
        a4 = self.create(ActionItem, UML.Action)
        jn = self.create(self.item_cls, self.join_node_cls)

        # connect actions first
        self.connect(flow1, flow1.head, a1)
        self.connect(flow2, flow2.tail, a2)
        self.connect(flow3, flow3.tail, a2)
        self.connect(flow4, flow4.head, a4)

        # connect to the node
        self.connect(flow1, flow1.tail, jn)
        self.connect(flow2, flow2.head, jn)
        self.connect(flow3, flow3.head, jn)
        self.connect(flow4, flow4.tail, jn)

        # needed for tests below
        cflow = jn.subject.outgoing[0]
        cnode = jn.combined
        assert cflow in self.kindof(UML.ControlFlow)
        assert cnode in self.kindof(self.fork_node_cls)

        # test disconnection
        self.disconnect(flow4, flow4.head)
        assert self.get_connected(flow4.head) is None
        assert jn.combined is None

        flows = self.kindof(UML.ControlFlow)
        nodes = self.kindof(self.fork_node_cls)
        assert cnode not in nodes, f"{cnode} in {nodes}"
        assert cflow not in flows, f"{cflow} in {flows}"


class FlowItemForkNodeTestCase(FlowItemDesisionAndForkNodes, TestCase):
    item_cls = ForkNodeItem
    fork_node_cls = UML.ForkNode
    join_node_cls = UML.JoinNode


class FlowItemDecisionNodeTestCase(FlowItemDesisionAndForkNodes, TestCase):
    item_cls = DecisionNodeItem
    fork_node_cls = UML.DecisionNode
    join_node_cls = UML.MergeNode



import pytest

from gaphor.services.eventmanager import EventManager
import gaphor.UML as UML


@pytest.fixture
def factory():
    event_manager = EventManager()
    return UML.ElementFactory(event_manager)


def test_association(factory):
    """Testing Association elements in the meta-model"""
    element = factory.create(UML.Association)

    property1 = factory.create(UML.Property)
    property2 = factory.create(UML.Property)

    element.memberEnd = property1
    element.memberEnd = property2

    element.ownedEnd = property1

    element.navigableOwnedEnd = property1

    assert (
        not element.isDerived
    ), f"The isDerived property should default to False - {element.isDerived}"
    assert (
        property1 in element.member
    ), f"Namespace.member does not contain memberEnd - {element.member}"
    assert (
        property2 in element.member
    ), f"Namespace.member does not contain memberEnd - {element.member}"

    assert (
        property1 in element.feature
    ), f"Classifier.feature does not contain ownedEnd - {element.feature}"
    assert (
        property1 in element.ownedMember
    ), f"Namespace.ownedMember does not contain ownedEnd - {element.ownedEnd}"

    assert property1 in element.ownedEnd, (
        "Association.ownedEnd does not contain navigableOwnedEnd - %s"
        % element.ownedEnd
    )


#    def test_association_class(self):
#        try:
#            element = self.factory.create(UML.AssociationClass)
#        except AttributeError:
#            self.fail('AssociationClass elements are not part of the meta-model')


def test_class(factory):
    """Testing Class elements in the meta-model"""
    element = factory.create(UML.Class)
    property1 = factory.create(UML.Property)
    operation1 = factory.create(UML.Operation)
    element.ownedAttribute = property1
    element.ownedOperation = operation1

    assert (
        property1 in element.attribute
    ), f"Classifier.attribute does not contain ownedAttribute - {element.attribute}"
    assert property1 in element.ownedMember, (
        "Namespace.ownedMember does not contain ownedAttribute - %s"
        % element.ownedMember
    )

    assert (
        operation1 in element.feature
    ), f"Classifier.feature does not contain ownedOperation - {element.feature}"
    assert operation1 in element.ownedMember, (
        "Namespace.ownedMember does not contain ownedOperation" % element.ownedMember
    )


def test_comment(factory):
    """Testing Comment elements in the meta-model"""
    element = factory.create(UML.Comment)
    element.body = "Comment body"
    annotatedElement = factory.create(UML.Class)
    element.annotatedElement = annotatedElement

    assert element.body == "Comment body", f"Incorrect comment body - {element.body}"
    assert (
        annotatedElement in element.annotatedElement
    ), f"Incorrect annotated element - {element.annotatedElement}"


def test_constraint(factory):
    """Testing Constraint elements in the meta-model"""
    element = factory.create(UML.Constraint)

    constrainedElement = factory.create(UML.Class)

    element.constrainedElement = constrainedElement
    element.specification = "Constraint specification"

    assert constrainedElement in element.constrainedElement, (
        "Constraint.constrainedElement does not contain the correct element - %s"
        % element.constrainedElement
    )
    assert (
        element.specification == "Constraint specification"
    ), f"Constraint.specification is incorrect - {element.specification}"


def test_dependency(factory):

    """Testing Dependency elements in the meta-model"""

    element = factory.create(UML.Dependency)

    client = factory.create(UML.Package)
    supplier = factory.create(UML.Package)

    element.client = client
    element.supplier = supplier

    assert (
        client in element.client
    ), f"Dependency.client does not contain client - {element.client}"
    assert (
        supplier in element.supplier
    ), f"Dependency.supplier does not contain supplier - {element.supplier}"


def test_element_import(factory):
    element = factory.create(UML.ElementImport)


def test_enumeration(factory):
    element = factory.create(UML.Enumeration)


def test_generalization(factory):
    element = factory.create(UML.Generalization)


def test_interface(factory):
    element = factory.create(UML.Interface)


def test_namespace(factory):
    element = factory.create(UML.Namespace)


def test_operation(factory):
    element = factory.create(UML.Operation)


def test_package(factory):
    element = factory.create(UML.Package)


def test_parameter(factory):
    element = factory.create(UML.Parameter)


def test_property(factory):
    element = factory.create(UML.Property)


def test_realization(factory):
    element = factory.create(UML.Realization)


def test_ids(factory):
    c = factory.create(UML.Class)
    assert c.id
    p = factory.create_as(UML.Class, id=False)
    assert p.id is False, p.id


def test1(factory):
    c = factory.create(UML.Class)
    p = factory.create(UML.Package)
    c.package = p
    assert c.package == p
    assert c.namespace == p
    assert c in p.ownedElement


def testOwnedMember_Unlink(factory):
    c = factory.create(UML.Class)
    p = factory.create(UML.Package)
    c.package = p

    c.unlink()

    assert [p] == factory.lselect()


#    def test_lower_upper(self):
#        """
#        Test MultiplicityElement.{lower|upper}
#        """
#        assert UML.MultiplicityElement.lowerValue in UML.MultiplicityElement.lower.subsets
#
#        e = UML.MultiplicityElement()
#        e.lowerValue = '2'
#        assert e.lower == '2', e.lower
#
#        assert UML.MultiplicityElement.upperValue in UML.MultiplicityElement.upper.subsets
#
#        e.upperValue = 'up'
#        assert UML.MultiplicityElement.upper.version == 4, UML.MultiplicityElement.upper.version
#        assert e.upper == 'up'
#        e.upperValue = 'down'
#        assert UML.MultiplicityElement.upper.version == 5, UML.MultiplicityElement.upper.version
#        assert e.upper == 'down', e.upper
#
#        # TODO: test signal handling


def test_property_is_composite(factory):
    p = UML.Property()
    assert p.isComposite == False, p.isComposite
    p.aggregation = "shared"
    assert p.isComposite == False, p.isComposite
    p.aggregation = "composite"
    assert p.isComposite == True, p.isComposite


def test_association_endType(factory):
    c1 = UML.Class()
    c2 = UML.Class()
    a = UML.Association()
    a.memberEnd = UML.Property()
    a.memberEnd = UML.Property()
    a.memberEnd[0].type = c1
    a.memberEnd[1].type = c2
    c1.ownedAttribute = a.memberEnd[0]
    c2.ownedAttribute = a.memberEnd[1]

    assert c1 in a.endType
    assert c2 in a.endType

    c3 = UML.Class()
    a.memberEnd[1].type = c3

    assert c1 in a.endType
    assert c3 in a.endType


def test_property_navigability(factory):
    p = factory.create(UML.Property)
    assert p.navigability is None

    c1 = factory.create(UML.Class)
    c2 = factory.create(UML.Class)
    a = UML.model.create_association(c1, c2)
    assert a.memberEnd[0].navigability is None
    assert a.memberEnd[1].navigability is None

    UML.model.set_navigability(a, a.memberEnd[0], True)
    assert a.memberEnd[0].navigability is True
    assert a.memberEnd[1].navigability is None

    UML.model.set_navigability(a, a.memberEnd[0], False)
    assert a.memberEnd[0].navigability is False
    assert a.memberEnd[1].navigability is None


def test_namedelement_qualifiedname(factory):
    p = factory.create(UML.Package)
    p.name = "Package"
    c = factory.create(UML.Class)
    c.name = "Class"

    assert ("Class",) == c.qualifiedName

    p.ownedClassifier = c

    assert ("Package", "Class") == c.qualifiedName


def test_extension_metaclass(factory):
    c = factory.create(UML.Class)
    c.name = "Class"
    s = factory.create(UML.Stereotype)
    s.name = "Stereotype"

    e = UML.model.create_extension(c, s)

    assert c == e.metaclass


def test_metaclass_extension(factory):
    c = factory.create(UML.Class)
    c.name = "Class"
    s = factory.create(UML.Stereotype)
    s.name = "Stereotype"

    assert [] == c.extension
    assert [] == s.extension

    e = UML.model.create_extension(c, s)

    print(e.memberEnd)
    assert [e] == c.extension
    assert [] == s.extension
    assert e.ownedEnd.type is s


def test_operation_parameter_deletion(factory):
    assert 0 == len(factory.lselect())

    c = factory.create(UML.Class)
    c.name = "Class"
    o = factory.create(UML.Operation)
    c.ownedOperation = o
    UML.parse(o, "a(x: int, y: int)")

    c.unlink()

    assert 0 == len(factory.lselect()), factory.lselect()



import math
from gi.repository import Gtk

from gaphor import UML
from gaphor.core import _, transactional
from gaphor.diagram.propertypages import (
    PropertyPages,
    NamedElementPropertyPage,
    NamedItemPropertyPage,
    create_hbox_label,
)
from gaphor.diagram.actions import ObjectNodeItem, ForkNodeItem


@PropertyPages.register(ObjectNodeItem)
class ObjectNodePropertyPage(NamedItemPropertyPage):
    """
    """

    ORDERING_VALUES = ["unordered", "ordered", "LIFO", "FIFO"]

    def construct(self):
        page = super().construct()

        subject = self.subject

        if not subject:
            return page

        hbox = create_hbox_label(self, page, _("Upper bound"))
        entry = Gtk.Entry()
        entry.set_text(subject.upperBound or "")
        entry.connect("changed", self._on_upper_bound_change)
        hbox.pack_start(entry, True, True, 0)

        hbox = create_hbox_label(self, page, "")
        combo = Gtk.ComboBoxText()
        for v in self.ORDERING_VALUES:
            combo.append_text(v)
        combo.set_active(self.ORDERING_VALUES.index(subject.ordering))
        combo.connect("changed", self._on_ordering_change)
        hbox.pack_start(combo, False, True, 0)

        hbox = create_hbox_label(self, page, "")
        button = Gtk.CheckButton(_("Ordering"))
        button.set_active(self.item.show_ordering)
        button.connect("toggled", self._on_ordering_show_change)
        hbox.pack_start(button, False, True, 0)

        return page

    def update(self):
        pass

    @transactional
    def _on_upper_bound_change(self, entry):
        value = entry.get_text().strip()
        self.subject.upperBound = value

    @transactional
    def _on_ordering_change(self, combo):
        value = self.ORDERING_VALUES[combo.get_active()]
        self.subject.ordering = value

    @transactional
    def _on_ordering_show_change(self, button):
        self.item.show_ordering = button.get_active()


@PropertyPages.register(ForkNodeItem)
class JoinNodePropertyPage(NamedItemPropertyPage):
    """
    """

    def construct(self):
        page = super().construct()

        subject = self.subject

        if not subject:
            return page

        hbox = Gtk.HBox()
        page.pack_start(hbox, False, True, 0)

        if isinstance(subject, UML.JoinNode):
            hbox = create_hbox_label(self, page, _("Join specification"))
            entry = Gtk.Entry()
            entry.set_text(subject.joinSpec or "")
            entry.connect("changed", self._on_join_spec_change)
            hbox.pack_start(entry, True, True, 0)

        button = Gtk.CheckButton(_("Horizontal"))
        button.set_active(self.item.matrix[2] != 0)
        button.connect("toggled", self._on_horizontal_change)
        page.pack_start(button, False, True, 0)

        return page

    def update(self):
        pass

    @transactional
    def _on_join_spec_change(self, entry):
        value = entry.get_text().strip()
        self.subject.joinSpec = value

    def _on_horizontal_change(self, button):
        if button.get_active():
            self.item.matrix.rotate(math.pi / 2)
        else:
            self.item.matrix.rotate(-math.pi / 2)
        self.item.request_update()


@PropertyPages.register(UML.ControlFlow)
@PropertyPages.register(UML.ObjectFlow)
class FlowPropertyPageAbstract(NamedElementPropertyPage):
    """Flow item element editor."""

    def construct(self):
        page = super().construct()

        subject = self.subject

        if not subject:
            return page

        hbox = create_hbox_label(self, page, _("Guard"))
        entry = Gtk.Entry()
        entry.set_text(subject.guard or "")
        changed_id = entry.connect("changed", self._on_guard_change)
        hbox.pack_start(entry, True, True, 0)

        def handler(event):
            entry.handler_block(changed_id)
            v = event.new_value
            entry.set_text(v if v else "")
            entry.handler_unblock(changed_id)

        self.watcher.watch("guard", handler).subscribe_all()
        entry.connect("destroy", self.watcher.unsubscribe_all)

        return page

    @transactional
    def _on_guard_change(self, entry):
        value = entry.get_text().strip()
        self.subject.guard = value



"""
For ease of creation, maintain a mapping from UML Element to Diagram Item.
"""

from typing import Dict
from gaphor import UML


def represents(uml_element):
    """
    A decorator to assign a default UML Element type to a diagram item.
    """

    def wrapper(presentation):
        set_diagram_item(uml_element, presentation)
        return presentation

    return wrapper


# Map UML elements to their (default) representation.
_uml_to_item_map: Dict[UML.Element, UML.Presentation] = {}


def get_diagram_item(element):
    global _uml_to_item_map
    return _uml_to_item_map.get(element)


def set_diagram_item(element, item):
    global _uml_to_item_map
    _uml_to_item_map[element] = item



"""
Test classes.
"""

from gaphor.tests import TestCase
from gaphor import UML
from gaphor.diagram.classes.interface import InterfaceItem, Folded


class InterfaceTestCase(TestCase):
    def test_interface_creation(self):
        """Test interface creation
        """
        iface = self.create(InterfaceItem, UML.Interface)
        assert isinstance(iface.subject, UML.Interface)

    def test_folded_interface_persistence(self):
        """Test folded interface saving/loading
        """
        iface = self.create(InterfaceItem, UML.Interface)

        # note: assembly folded mode..
        iface.folded = Folded.REQUIRED

        data = self.save()
        self.load(data)

        interfaces = self.diagram.canvas.select(lambda e: isinstance(e, InterfaceItem))
        assert 1 == len(interfaces)
        # ... gives provided folded mode on load;
        # correct folded mode is determined by connections, which will be
        # recreated later, i.e. required folded mode will be set when
        # implementation connects to the interface
        # assert Folded.PROVIDED == interfaces[0].folded



from gaphor.ui.iconname import to_kebab_case


def test_kebab_case():
    assert to_kebab_case("Class") == "class"
    assert to_kebab_case("FooBar") == "foo-bar"
    assert to_kebab_case("FooBarBaz") == "foo-bar-baz"



"""
Test state items.
"""

from gaphor import UML
from gaphor.diagram.states.state import StateItem
from gaphor.tests.testcase import TestCase


class StateTestCase(TestCase):
    def test_state(self):
        """Test creation of states
        """
        self.create(StateItem, UML.State)

    def test_activities_persistence(self):
        """Test state activities saving/loading
        """
        # all activities
        s1 = self.create(StateItem, UML.State)
        s1.subject.name = "s1"
        s1.subject.entry = s1.model.create(UML.Activity)
        s1.subject.exit = s1.model.create(UML.Activity)
        s1.subject.doActivity = s1.model.create(UML.Activity)
        s1.subject.entry.name = "test 1 entry"
        s1.subject.exit.name = "test 1 exit"
        s1.subject.doActivity.name = "test 1 do"

        data = self.save()
        self.load(data)

        states = self.diagram.canvas.select(lambda e: isinstance(e, StateItem))
        assert 1 == len(states)
        s1 = states[0]

        assert "test 1 entry" == s1.subject.entry.name
        assert "test 1 exit" == s1.subject.exit.name
        assert "test 1 do" == s1.subject.doActivity.name



"""
Component item.
"""

from gaphor import UML
from gaphor.diagram.classes.stereotype import stereotype_compartments
from gaphor.diagram.presentation import ElementPresentation, Classified
from gaphor.diagram.shapes import Box, EditableText, Text, draw_border
from gaphor.diagram.text import VerticalAlign, FontWeight
from gaphor.diagram.support import represents


@represents(UML.Component)
class ComponentItem(ElementPresentation, Classified):
    def __init__(self, id=None, model=None):
        super().__init__(id, model)

        self.watch("show_stereotypes", self.update_shapes)
        self.watch("subject[NamedElement].name")
        self.watch("subject.appliedStereotype", self.update_shapes)
        self.watch("subject.appliedStereotype.classifier.name")
        self.watch("subject.appliedStereotype.slot", self.update_shapes)
        self.watch("subject.appliedStereotype.slot.definingFeature.name")
        self.watch("subject.appliedStereotype.slot.value", self.update_shapes)

    show_stereotypes = UML.properties.attribute("show_stereotypes", int)

    def update_shapes(self, event=None):
        self.shape = Box(
            Box(
                Text(
                    text=lambda: UML.model.stereotypes_str(self.subject),
                    style={"min-width": 0, "min-height": 0},
                ),
                EditableText(
                    text=lambda: self.subject.name or "",
                    style={"font-weight": FontWeight.BOLD},
                ),
                style={"padding": (4, 34, 4, 4), "min-height": 44},
                draw=draw_component_icon,
            ),
            *(self.show_stereotypes and stereotype_compartments(self.subject) or []),
            style={
                "min-width": 100,
                "min-height": 50,
                "vertical-align": VerticalAlign.TOP,
            },
            draw=draw_border
        )


def draw_component_icon(box, context, bounding_box):
    bar_width = 12
    bar_height = 4
    bar_padding = 4
    icon_width = 16
    icon_height = 20
    icon_margin_x = 6
    icon_margin_y = 12

    cr = context.cairo

    ix = bounding_box.width - icon_margin_x - icon_width
    iy = icon_margin_y

    cr.save()
    try:
        cr.set_line_width(1.0)
        cr.rectangle(ix, iy, icon_width, icon_height)
        cr.stroke()

        bx = ix - bar_padding
        bar_upper_y = iy + bar_padding
        bar_lower_y = iy + bar_padding * 3

        color = cr.get_source()
        cr.rectangle(bx, bar_lower_y, bar_width, bar_height)
        cr.set_source_rgb(1, 1, 1)  # white
        cr.fill_preserve()
        cr.set_source(color)
        cr.stroke()

        cr.rectangle(bx, bar_upper_y, bar_width, bar_height)
        cr.set_source_rgb(1, 1, 1)  # white
        cr.fill_preserve()
        cr.set_source(color)
        cr.stroke()
    finally:
        cr.restore()



"""
Trivial drawing aids (box, line, ellipse).
"""

import ast

from gaphas.item import Element, NW
from gaphas.item import Line as _Line
from gaphas.util import path_ellipse


class Line(_Line):
    def __init__(self, id=None, model=None):
        super().__init__()
        self.style = {"line-width": 2, "color": (0, 0, 0, 1)}.__getitem__
        self._id = id
        self.fuzziness = 2
        self._handles[0].connectable = False
        self._handles[-1].connectable = False

    id = property(lambda self: self._id, doc="Id")

    def save(self, save_func):
        save_func("matrix", tuple(self.matrix))
        for prop in ("orthogonal", "horizontal"):
            save_func(prop, getattr(self, prop))
        points = [tuple(map(float, h.pos)) for h in self.handles()]
        save_func("points", points)

    def load(self, name, value):
        if name == "matrix":
            self.matrix = ast.literal_eval(value)
        elif name == "points":
            points = ast.literal_eval(value)
            for x in range(len(points) - 2):
                h = self._create_handle((0, 0))
                self._handles.insert(1, h)
            for i, p in enumerate(points):
                self.handles()[i].pos = p
            self._update_ports()
        elif name == "horizontal":
            self.horizontal = ast.literal_eval(value)
        elif name == "orthogonal":
            self._load_orthogonal = ast.literal_eval(value)

    def postload(self):
        if hasattr(self, "_load_orthogonal"):
            self.orthogonal = self._load_orthogonal
            del self._load_orthogonal

    def draw(self, context):
        cr = context.cairo
        style = self.style
        cr.set_line_width(style("line-width"))
        cr.set_source_rgba(*style("color"))
        super().draw(context)


class Box(Element):
    """
    A Box has 4 handles (for a start)::

    NW +---+ NE
    SW +---+ SE
    """

    def __init__(self, id=None, model=None):
        super().__init__(10, 10)
        self.style = {"line-width": 2, "color": (0, 0, 0, 1)}.__getitem__
        self._id = id

    id = property(lambda self: self._id, doc="Id")

    def save(self, save_func):
        save_func("matrix", tuple(self.matrix))
        save_func("width", self.width)
        save_func("height", self.height)

    def load(self, name, value):
        if name == "matrix":
            self.matrix = ast.literal_eval(value)
        elif name == "width":
            self.width = ast.literal_eval(value)
        elif name == "height":
            self.height = ast.literal_eval(value)

    def postload(self):
        pass

    def draw(self, context):
        cr = context.cairo
        nw = self._handles[NW]
        style = self.style
        cr.rectangle(nw.pos.x, nw.pos.y, self.width, self.height)
        # cr.set_source_rgba(*style("color"))
        # cr.fill_preserve()
        cr.set_source_rgba(*style("color"))
        cr.set_line_width(style("line-width"))
        cr.stroke()


class Ellipse(Element):
    """
    """

    def __init__(self, id=None, model=None):
        super().__init__()
        self.style = {"line-width": 2, "color": (0, 0, 0, 1)}.__getitem__
        self._id = id

    id = property(lambda self: self._id, doc="Id")

    def save(self, save_func):
        save_func("matrix", tuple(self.matrix))
        save_func("width", self.width)
        save_func("height", self.height)

    def load(self, name, value):
        if name == "matrix":
            self.matrix = ast.literal_eval(value)
        elif name == "width":
            self.width = ast.literal_eval(value)
        elif name == "height":
            self.height = ast.literal_eval(value)

    def postload(self):
        pass

    def draw(self, context):
        cr = context.cairo
        nw = self._handles[NW]
        style = self.style

        rx = self.width / 2.0
        ry = self.height / 2.0

        cr.move_to(self.width, ry)
        path_ellipse(cr, rx, ry, self.width, self.height)
        # cr.set_source_rgba(*style.fill_color)
        # cr.fill_preserve()
        cr.set_source_rgba(*style("color"))
        cr.set_line_width(style("line-width"))
        cr.stroke()



"""
Toolbox.
"""


import logging

from gi.repository import GLib, Gdk, Gtk

from gaphor.core import _
from gaphor.abc import ActionProvider
from gaphor.ui.abc import UIComponent
from gaphor.ui.diagramtoolbox import TOOLBOX_ACTIONS

log = logging.getLogger(__name__)


class Toolbox(UIComponent, ActionProvider):

    TARGET_STRING = 0
    TARGET_TOOLBOX_ACTION = 1
    DND_TARGETS = [
        Gtk.TargetEntry.new("STRING", Gtk.TargetFlags.SAME_APP, TARGET_STRING),
        Gtk.TargetEntry.new("text/plain", Gtk.TargetFlags.SAME_APP, TARGET_STRING),
        Gtk.TargetEntry.new(
            "gaphor/toolbox-action", Gtk.TargetFlags.SAME_APP, TARGET_TOOLBOX_ACTION
        ),
    ]

    title = _("Toolbox")

    def __init__(self, main_window, properties, toolbox_actions=TOOLBOX_ACTIONS):
        self.main_window = main_window
        self.properties = properties
        self._toolbox = None
        self._toolbox_actions = toolbox_actions

    def open(self):
        widget = self.construct()

        return widget

    def close(self):
        if self._toolbox:
            self._toolbox.destroy()
            self._toolbox = None

    def construct(self):
        def toolbox_button(action_name, icon_name, label, shortcut):
            button = Gtk.ToggleToolButton.new()
            icon = Gtk.Image.new_from_icon_name(icon_name, Gtk.IconSize.BUTTON)
            button.set_icon_widget(icon)
            button.set_action_name("diagram.select-tool")
            button.set_action_target_value(GLib.Variant.new_string(action_name))
            if label:
                if shortcut:
                    a, m = Gtk.accelerator_parse(shortcut)
                    button.set_tooltip_text(
                        f"{label} ({Gtk.accelerator_get_label(a, m)})"
                    )
                else:
                    button.set_tooltip_text(f"{label}")

            # Enable Drag and Drop
            if action_name != "toolbox-pointer":
                inner_button = button.get_children()[0]
                inner_button.drag_source_set(
                    Gdk.ModifierType.BUTTON1_MASK | Gdk.ModifierType.BUTTON3_MASK,
                    self.DND_TARGETS,
                    Gdk.DragAction.COPY | Gdk.DragAction.LINK,
                )
                inner_button.drag_source_set_icon_name(icon_name)
                inner_button.connect(
                    "drag-data-get", self._button_drag_data_get, action_name
                )

            return button

        toolbox = Gtk.ToolPalette.new()
        toolbox.connect("destroy", self._on_toolbox_destroyed)

        collapsed = self.properties.get("toolbox-collapsed", {})

        def on_collapsed(widget, prop, index):
            collapsed[index] = widget.get_property("collapsed")
            self.properties.set("toolbox-collapsed", collapsed)

        for index, (title, items) in enumerate(self._toolbox_actions):
            tool_item_group = Gtk.ToolItemGroup.new(title)
            tool_item_group.set_property("collapsed", collapsed.get(index, False))
            tool_item_group.connect("notify::collapsed", on_collapsed, index)
            for action_name, label, icon_name, shortcut in items:
                button = toolbox_button(action_name, icon_name, label, shortcut)
                tool_item_group.insert(button, -1)
                button.show_all()

            toolbox.add(tool_item_group)
            tool_item_group.show()

        toolbox.show()

        self._toolbox = toolbox

        scrolled_window = Gtk.ScrolledWindow()
        scrolled_window.set_policy(Gtk.PolicyType.NEVER, Gtk.PolicyType.AUTOMATIC)
        scrolled_window.set_shadow_type(Gtk.ShadowType.IN)
        scrolled_window.add(toolbox)
        scrolled_window.show()
        return scrolled_window

    def _on_key_press_event(self, view, event):
        """
        Grab top level window events and select the appropriate tool based on the event.
        """
        if event.get_state() & Gdk.ModifierType.SHIFT_MASK or (
            event.get_state() == 0 or event.get_state() & Gdk.ModifierType.MOD2_MASK
        ):
            keyval = Gdk.keyval_name(event.keyval)
            self.set_active_tool(shortcut=keyval)

    def _on_toolbox_destroyed(self, widget):
        self._toolbox = None

    def set_active_tool(self, action_name=None, shortcut=None):
        """
        Set the tool based on the name of the action
        """
        # HACK:
        toolbox = self._toolbox
        if shortcut and toolbox:
            action_name = self.shortcuts.get(shortcut)
            log.debug(f"Action for shortcut {shortcut}: {action_name}")
            if not action_name:
                return

    def _button_drag_data_get(self, button, context, data, info, time, action_name):
        """The drag-data-get event signal handler.

        The drag-data-get signal is emitted on the drag source when the drop
        site requests the data which is dragged.

        Args:
            button (Gtk.Button): The button that received the signal.
            context (Gdk.DragContext): The drag context.
            data (Gtk.SelectionData): The data to be filled with the dragged
                data.
            info (int): The info that has been registered with the target in
                the Gtk.TargetList
            time (int): The timestamp at which the data was received.

        """
        data.set(type=data.get_target(), format=8, data=action_name.encode())



from gaphor.diagram.classes.association import AssociationItem
from gaphor.diagram.classes.dependency import DependencyItem
from gaphor.diagram.classes.generalization import GeneralizationItem
from gaphor.diagram.classes.implementation import ImplementationItem
from gaphor.diagram.classes.interface import InterfaceItem, Folded
from gaphor.diagram.classes.klass import ClassItem
from gaphor.diagram.classes.package import PackageItem


def _load():
    from gaphor.diagram.classes import (
        classconnect,
        interfaceconnect,
        classeseditors,
        classespropertypages,
    )


_load()



"""
The Sanitize module is dedicated to adapters (stuff) that keeps
the model clean and in sync with diagrams.
"""


from gaphor import UML
from gaphor.UML.event import AssociationDeleted, AssociationSet
from gaphor.core import event_handler
from gaphor.abc import Service


class SanitizerService(Service):
    """
    Does some background cleanup jobs, such as removing elements from the
    model that have no presentations (and should have some).
    """

    def __init__(self, event_manager):
        self.event_manager = event_manager

        event_manager.subscribe(self._unlink_on_presentation_delete)
        event_manager.subscribe(self._unlink_on_stereotype_delete)
        event_manager.subscribe(self._unlink_on_extension_delete)
        event_manager.subscribe(self._disconnect_extension_end)

    def shutdown(self):
        self.event_manager.unsubscribe(self._unlink_on_presentation_delete)
        self.event_manager.unsubscribe(self._unlink_on_stereotype_delete)
        self.event_manager.unsubscribe(self._unlink_on_extension_delete)
        self.event_manager.unsubscribe(self._disconnect_extension_end)

    @event_handler(AssociationDeleted)
    def _unlink_on_presentation_delete(self, event):
        """
        Unlink the model element if no more presentations link to the `item`'s
        subject or the deleted item was the only item currently linked.
        """
        if event.property is UML.Element.presentation:
            old_presentation = event.old_value
            if old_presentation and not event.element.presentation:
                event.element.unlink()

    def perform_unlink_for_instances(self, st, meta):

        inst = UML.model.find_instances(st)

        for i in list(inst):
            for e in i.extended:
                if not meta or isinstance(e, meta):
                    i.unlink()

    @event_handler(AssociationDeleted)
    def _unlink_on_extension_delete(self, event):
        """
        Remove applied stereotypes when extension is deleted.
        """
        if (
            isinstance(event.element, UML.Extension)
            and event.property is UML.Association.memberEnd
            and event.element.memberEnd
        ):
            p = event.element.memberEnd[0]
            ext = event.old_value
            if isinstance(p, UML.ExtensionEnd):
                p, ext = ext, p
            st = ext.type
            meta = p.type and getattr(UML, p.type.name, None)
            if st:
                self.perform_unlink_for_instances(st, meta)

    @event_handler(AssociationSet)
    def _disconnect_extension_end(self, event):
        if event.property is UML.ExtensionEnd.type and event.old_value:
            ext = event.element
            p = ext.opposite
            if not p:
                return
            st = event.old_value
            meta = getattr(UML, p.type.name, None)
            if st:
                self.perform_unlink_for_instances(st, meta)

    @event_handler(AssociationDeleted)
    def _unlink_on_stereotype_delete(self, event):
        """
        Remove applied stereotypes when stereotype is deleted.
        """
        if event.property is UML.InstanceSpecification.classifier:
            if isinstance(event.old_value, UML.Stereotype):
                event.element.unlink()



"""The core UML metamodel events."""

from gaphor.event import ServiceEvent


class ElementUpdated:
    """
    Generic event fired when element state changes.
    """

    def __init__(self, element, property):
        self.element = element
        self.property = property


class AttributeUpdated(ElementUpdated):
    """A UML attribute has changed value."""

    def __init__(self, element, attribute, old_value, new_value):
        """Constructor.  The element parameter is the element with the
        changing attribute.  The attribute parameter is the parameter
        element that changed.  The old_value is the old value of the attribute
        and the new_value is the new value of the attribute."""

        super().__init__(element, attribute)
        self.old_value = old_value
        self.new_value = new_value


class AssociationUpdated(ElementUpdated):
    """An association UML element has changed."""

    def __init__(self, element, association):
        """Constructor.  The element parameter is the element the association
        is changing from.  The association parameter is the changed
        association element."""

        super().__init__(element, association)


class AssociationSet(AssociationUpdated):
    """An association element has been set."""

    def __init__(self, element, association, old_value, new_value):
        """Constructor.  The element parameter is the element setting the
        association element.  The association parameter is the association
        element being set.  The old_value parameter is the old association
        and the new_value parameter is the new association."""

        AssociationUpdated.__init__(self, element, association)
        self.old_value = old_value
        self.new_value = new_value


class AssociationAdded(AssociationUpdated):
    """An association element has been added."""

    def __init__(self, element, association, new_value):
        """Constructor.  The element parameter is the element the association
        has been added to.  The association parameter is the association
        element being added."""

        AssociationUpdated.__init__(self, element, association)
        self.new_value = new_value


class AssociationDeleted(AssociationUpdated):
    """An association element has been deleted."""

    def __init__(self, element, association, old_value):
        """Constructor.  The element parameter is the element the association
        has been deleted from.  The association parameter is the deleted
        association element."""

        AssociationUpdated.__init__(self, element, association)
        self.old_value = old_value


class DerivedUpdated(AssociationUpdated):
    """A derived property has changed."""


class DerivedSet(AssociationSet, DerivedUpdated):
    """A generic derived set event."""

    def __init__(self, element, association, old_value, new_value):
        """Constructor.  The element parameter is the element to which the
        derived set belongs.  The association parameter is the association
        of the derived set."""

        super().__init__(element, association, old_value, new_value)


class DerivedAdded(AssociationAdded, DerivedUpdated):
    """A derived property has been added."""

    def __init__(self, element, association, new_value):
        """Constructor.  The element parameter is the element to which the
        derived property belongs.  The association parameter is the
        association of the derived property."""

        super().__init__(element, association, new_value)


class DerivedDeleted(AssociationDeleted, DerivedUpdated):
    """A derived property has been deleted."""

    def __init__(self, element, association, old_value):
        """Constructor.  The element parameter is the element to which the
        derived property belongs.  The association parameter is the
        association of the derived property."""

        super().__init__(element, association, old_value)


class RedefinedSet(AssociationSet):
    """A redefined property has been set."""

    def __init__(self, element, association, old_value, new_value):
        """Constructor.  The element parameter is the element to which the
        property belongs.  The association parameter is association of the
        property."""

        super().__init__(element, association, old_value, new_value)


class RedefinedAdded(AssociationAdded):
    """A redefined property has been added."""

    def __init__(self, element, association, new_value):
        """Constructor.  The element parameter is the element to which the
        property belongs.  The association parameter is the association of
        the property."""

        super().__init__(element, association, new_value)


class RedefinedDeleted(AssociationDeleted):
    """A redefined property has been deleted."""

    def __init__(self, element, association, old_value):
        """Constructor.  The element parameter is the element to which the
        property belongs.  The association parameter is the association of
        the property."""

        super().__init__(element, association, old_value)


class ElementCreated(ServiceEvent):
    """An element has been created."""

    def __init__(self, service, element):
        """Constructor.  The service parameter is the service responsible
        for creating the element.  The element parameter is the element being
        created."""
        super().__init__(service)
        self.element = element


class DiagramItemCreated(ElementCreated):
    """A diagram item has been created."""

    def __init__(self, service, element):
        """Constructor.  The element parameter is the element being created."""
        super().__init__(service, element)


class ElementDeleted(ServiceEvent):
    """An element has been deleted."""

    def __init__(self, service, element):
        """Constructor.  The service parameter is the service responsible for
        deleting the element.  The element parameter is the element being
        deleted."""
        super().__init__(service)
        self.element = element


class ModelReady(ServiceEvent):
    """A generic element factory event."""

    def __init__(self, service):
        """Constructor.  The service parameter is the service the emitted the
        event."""
        super().__init__(service)


class ModelFlushed(ServiceEvent):
    """The element factory has been flushed."""

    def __init__(self, service):
        """Constructor.  The service parameter is the service responsible for
        flushing the factory."""
        super().__init__(service)
